#!/usr/bin/python

# Compute the posterior uncertainty covariance.


import sys
import numpy
import getopt
import datetime

sys.path.append("../src")
from utils import *
#from scipy.sparse import csr_matrix


#########################################################################
def usage():

        print """
        apost.py [-c configfile] [-b] sigmafile

	where 
		sigmafile : file with sprior sigma values.
		-c        : Specify the configuration file to use.
		-b        : If set, use boundary value optimization files.

        """


####################################################################
def getHQSum(config, start_ts, end_ts):
	""" Read in the saved hq files over certain time steps and compute sum """

	hqsum = numpy.zeros( (config["num_obs"], config["nlandcells"]) )

        t0 = datetime.datetime.now()
        for i in range(start_ts, end_ts):
                t1 = datetime.datetime.now()

                # hq slice files are numbered 1 .. ntimesteps
                # hq shape is #obs x #cells
                hqfile = config["hqdir"] + "/HQ%04d.npy" % (i+1)
                hq = numpy.load(hqfile)
                hqsum = hqsum + hq

                t2a = datetime.datetime.now()
                print "Step %4d" % (i), "of ", start_ts, "to", end_ts-1, "done. Step time ", t2a-t1, "Total elapsed time ", t2a-t0

	return hqsum


####################################################################
def getQSum(start_ts, end_ts, sigma, D, E):

	# Qsum is (S^T * D * S) * E 
	# where
	#   S is time-space varying sigma (#timesteps x # cells)
	#   D is temporal covariance (#timesteps x # timesteps)
	#   S^T is transpose of S (#cells x #timesteps)
	#   E is spatial covariance (#cells x #cells)

	x = numpy.dot(sigma[start_ts:end_ts].T, D[start_ts:end_ts,start_ts:end_ts])

	# multiplication with E is an element by element multiply, NOT matrix multiply
	# In this case E is a numpy array, so the * will do element multiply
	qsum = numpy.dot(x, sigma[start_ts:end_ts]) * E

	return qsum


####################################################################
def mkvnc(filename, name, v, description=None):
	""" Create the netcdf file for given v array """

	ds = netCDF4.Dataset(filename, 'w', format='NETCDF4')

	size = v.shape[0]

	ds.createDimension('size', size)
	grid = ds.createVariable(name, 'double', ('size', 'size'))
	if description is None:
		grid.description = "aggregated posterior uncertainty covariance"
	else:
		grid.description = description
	grid[:] = v

	ds.close()


####################################################################
def dateToTimestep(config, date):
	""" Calculate timestep number for a given date """

	(syr, smon, sday) = date.split("-")
	ddate = datetime.datetime(int(syr), int(smon), int(sday))
	td = ddate - config["sdate"]
	ts = td.days * config["steps_per_day"]

	return int(ts)


####################################################################
def getSigma(config, sigmafile):

	try:
		sigma = numpy.load(sigmafile)
	except:
		try:
			landmask = numpy.load(config["landmask_file"])
			sigma = getSpriorNetcdf(sigmafile, "sigma", landmask)
		except:
			print "Cannot read sigma file."
			sys.exit()

	return sigma


####################################################################
def make_vshat(start_ts, end_ts, config, TC, E, hqhti, sigmafile, include_boundary, start_ts_bound=None, end_ts_bound=None):
	""" Calculate vshat for the timesteps from start_ts to end_ts. """


	# get hqsum for flux
	hqsum = getHQSum(config, start_ts, end_ts)
	print "hqsum shape is ", hqsum.shape

	# get hqsum for boundary levels
	if include_boundary:
                config_pbl = getConfig("config_pbl.ini")
                config_trans = getConfig("config_trans.ini")
                config_ftrop = getConfig("config_ftrop.ini")

		hqsum_pbl   = getHQSum(config_pbl, start_ts_bound, end_ts_bound)
		hqsum_trans = getHQSum(config_trans, start_ts_bound, end_ts_bound)
		hqsum_ftrop = getHQSum(config_ftrop, start_ts_bound, end_ts_bound)
		hqsum = numpy.hstack( (hqsum, hqsum_pbl, hqsum_trans, hqsum_ftrop) )
		print "hqsum with boundary shape is ", hqsum.shape


	print "compute qsum for flux"
	# get time, space varying sigma from saved file
	sigma = getSigma(config, sigmafile)
	qsum = getQSum(start_ts, end_ts, sigma, TC, E)

	if include_boundary:
		print "compute qsum for boundary levels"

		# need some way to get sigma file names for each boundary level
		sigma_pbl = getSigma(config_pbl, "sigma_pbl.npy")
		sigma_trans = getSigma(config_trans, "sigma_trans.npy")
		sigma_ftrop = getSigma(config_ftrop, "sigma_ftrop.npy")

		pbl_spatial_cl = config_pbl["spatial_corr_length"]
		trans_spatial_cl = config_trans["spatial_corr_length"]
		ftrop_spatial_cl = config_ftrop["spatial_corr_length"]

		# calculate E for each boundary
		heff_bound = numpy.load(config_pbl["sp_cov_file"])
		Epbl = numpy.exp(-heff_bound/pbl_spatial_cl)
		Etrans = numpy.exp(-heff_bound/trans_spatial_cl)
		Eftrop = numpy.exp(-heff_bound/ftrop_spatial_cl)

		qsum_pbl   = getQSum(start_ts_bound, end_ts_bound, sigma_pbl, TC, Epbl)
		qsum_trans = getQSum(start_ts_bound, end_ts_bound, sigma_trans, TC, Etrans)
		qsum_ftrop = getQSum(start_ts_bound, end_ts_bound, sigma_ftrop, TC, Eftrop)

		qsum = block_diag(qsum, qsum_pbl, qsum_trans, qsum_ftrop)


	print "compute vshat"
	a = numpy.dot(hqsum.T,  hqhti)
	a = numpy.dot(a, hqsum)
	vshat = (qsum - a) / (end_ts - start_ts)**2

	filename = "qsum_%d-%d.nc" % (start_ts, end_ts)
	varname = "qsum"
	mkvnc(filename, varname, qsum, "Sum of Q matrix between time steps %s and %s" % (start_ts, end_ts))

	return vshat


####################################################################


if __name__ == '__main__':


	numpy.set_printoptions(precision=2, edgeitems=10, linewidth=240)

	configfile = "config.ini"
	include_boundary = False

	try:
		opts, args = getopt.getopt(sys.argv[1:], "bc:")
	except getopt.GetoptError, err:
		# print help information and exit:
		print str(err) # will print something like "option -a not recognized"
		usage()
		sys.exit()

	for o, a in opts:
		if o == "-b":
			include_boundary = True
		if o == "-c":
			configfile = a

	if len(args) != 1:
		usage()
		sys.exit()

	try:
		sigmafile = args[0]
	except:
		usage()
		sys.exit()

	config = getConfig(configfile)

	tstart = datetime.datetime.now()

	# Get temporal covariance matrix
	print "Calculating temporal covariance..."
	temporal_cl = config["temporal_corr_length"]
	TC = makeTemporalCovariance(temporal_cl, config)

	# Get spatial covariance matrix.
	# E is exp(-Xs/ls) where Xs is distance between cells, ls is spatial correlation parameter
	print "Loading spatial covariance..."
	heff = numpy.load(config["sp_cov_file"])	# distance matrix
	spatial_cl = config["spatial_corr_length"]
	E = numpy.exp(-heff/spatial_cl)

	# get the (hqht +r)^-1 computed from inversion
	print "Reading hqhti.npy..."
	hqhti = numpy.load("hqhti.npy")
	num_obs = hqhti.shape[0]
	config["num_obs"] = num_obs



	# hard coded dates for aggregated covariance
	covar_startdates = config["vshat_dates"]
	for i in range(len(covar_startdates)-1):
		print i, covar_startdates[i]

		# convert date to timestep number
		start_ts = dateToTimestep(config, covar_startdates[i])
		end_ts = dateToTimestep(config, covar_startdates[i+1])
		print start_ts, end_ts

		if include_boundary:
			config_pbl = getConfig("config_pbl.ini")
			start_ts_bound = dateToTimestep(config_pbl, covar_startdates[i])
			end_ts_bound = dateToTimestep(config_pbl, covar_startdates[i+1])
			vshat = make_vshat(start_ts, end_ts, config, TC, E, hqhti, sigmafile, include_boundary, start_ts_bound, end_ts_bound)
		else:
			vshat = make_vshat(start_ts, end_ts, config, TC, E, hqhti, sigmafile, include_boundary)

		filename = "vshat_%d-%d" % (start_ts, end_ts)
		numpy.save(filename, vshat)
		mkvnc(filename + ".nc", "vshat", vshat)


	# now redo for entire time span, because we can't be certain that
	# it was covered by config['vshat_dates']
	start_ts = dateToTimestep(config, config["start_date"])
	end_ts = dateToTimestep(config, config["end_date"])
	if include_boundary:
		config_pbl = getConfig("config_pbl.ini")
		start_ts_bound = dateToTimestep(config_pbl, covar_startdates[i])
		end_ts_bound = dateToTimestep(config_pbl, covar_startdates[i+1])
		vshat = make_vshat(start_ts, end_ts, config, TC, E, hqhti, sigmafile, include_boundary, start_ts_bound, end_ts_bound)
	else:
		vshat = make_vshat(start_ts, end_ts, config, TC, E, hqhti, sigmafile, include_boundary)

	numpy.save("vshat", vshat)
	mkvnc("vshat.nc", "vshat", vshat)
#	numpy.save("hqsum_a", hqsum_a)

	tend = datetime.datetime.now()
	print "Total elapsed time is", tend - tstart
